/***************************************************************************
 * The contents of this file are subject to the NOKOS License Version 1.0a
 * (the "License"); you may not use this file except in compliance with the
 * License. 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF  ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The Original Software is Nokia D211 Software for Linux.
 *
 * Copyright (c) 2002 Nokia. All Rights Reserved.
 *
 * Nokia is a registered trademark of Nokia Corporation.
 * Other product and company names mentioned herein may be
 * trademarks or tradenames of their respective owners. 
 *
 * Contributor(s): ______________________________________. 
 *
 **************************************************************************/

#include <string.h>
#include "naap_am.h"
#include "am_fileconfig_utils.h"
#include "naap_dt.h"

#define NAAP_LINE_LENGTH 2048

static char* get_subsect (FILE* file)
{
	static char line [NAAP_LINE_LENGTH];
	char* result;
	
	result = fgets (line, sizeof line, file);
	
	if (result && *line == '\n'){
		result = NULL;
	}
	
	return result;
}

static int is_sect (char* line, char const* sectname)
{
	size_t length;
	
	if (strlen (line) < (length = strlen (sectname))){
		return 0;
	}
	
	return strncmp (sectname, line, length) == 0;
}

void cnf_subsects (FILE* file, SubSection* subsects, void* data)
{
	SubSection* iter;
	char* line;
	int found;
	
	for (line = get_subsect (file); line; line = get_subsect (file)) {
		for (found = 0, iter = subsects; iter->name; iter++) {
			if (is_sect (line, iter->name)) {
				line [strlen (line) - 1] = '\0';
				(*iter->handler) (line, data);
				found = 1;
				break;
			}
		}
		if (!found) {
			DT((DT_CONF, "ERROR: cnf_subsects: unrecognized subsection: %s", line));
		}
	}
}

static int find_section (FILE* file, Section* section)
{
	char line [NAAP_LINE_LENGTH];
	
	while (fgets (line, sizeof (line), file)){
		if (is_sect (line, section->name)){
			return 1;
		}
	}
	
	return 0;
}

static void cnf_sect (FILE* file, Section* section)
{
	rewind (file);
	
	while (find_section (file, section)){
		(*section->handler) (file);
	}
}

static char* find_rvalue (char* param)
{
	char* rvalue;
	
	for (rvalue = param; *rvalue; rvalue++){
		if (*rvalue == '='){
			break;
		}
	}
	
	if (!*rvalue){
             DT((DT_AM, "incorrect assignment statement: %s\n", param));
             return NULL;
	}
	
	for (rvalue++; *rvalue == ' '; rvalue++){
		/* empty */
	}
	
	return rvalue;
}

void cnf_get_string (char* target, char* param, size_t maxlen)
{
	size_t len;
	
	param = find_rvalue (param);
	
	if (!param){
		return;
	}
	
	len = min (maxlen - 1, strlen (param));
	memcpy (target, param, len);
	target[len] = '\0';
}

IP_addr cnf_get_addr (char* param)
{
	IP_addr addr;
	
	param = find_rvalue (param);
	
	if (!param){
             return am_htonl (IP_ADDR_ANY);
	}

	if (((addr = am_inet_aton (param)) == IP_ADDR_NONE) && (strcmp("255.255.255.255", param))){
             DT((DT_AM, "incorrect address in assignment: %s\n", param));
             return am_htonl (IP_ADDR_ANY);
	}
	
	return addr;
}

u_int32_t cnf_get_int32 (char* param)
{
	param = find_rvalue (param);
	
	if (!param){
		return 0;
	}
	
	return atol (param);
}

int cnf_get_bool (char* param)
{
	param = find_rvalue (param);
	
	if (!param){
		return 0;
	}
	
	if (*param == '1'){
		return 1;
	}
	
	if (*param == 'y'){
		return 1;
	}
	
	return 0;
}

int cnf_get_enum (char* param, char const** enums)
{
	size_t paramlen;
	int iter;
	
	param = find_rvalue (param);
	
	if (!param){
		return 0;
	}
	
	paramlen = strlen (param);
	
	for (iter = 0; enums [iter]; iter++){
		size_t enumlen;
		
		if (paramlen < (enumlen = strlen (enums [iter]))) {
			continue;
		}
		
		if (strncmp (param, enums [iter], enumlen) == 0) {
			return iter;
		}
	}
	
	DT((DT_AM, "unrecognized enum: %s\n", param));
	return 0;
}

void cnf_get_hex_buf (char* line, u_int8_t* buffer, u_int16_t *buffersize)
{
	char* next = NULL;
	int32_t result;
	size_t length = 0;
	
	line = find_rvalue (line);
	
	if (!line){
		*buffersize = 0;
		return;
	}
	
	for (;;){
		result = strtol (line, &next, 16);
		
		if (next == line){
                     DT((DT_AM, "unrecognized hex number: %s\n", line));
                     *buffersize = 0;
                     return;
		}
		
		if (result < 0 || result > 255){
                     DT((DT_AM, "hex number overflow, valid range 0x00-0xff: %s\n", line));
                     *buffersize = 0;
                     return;
		}
		
		if (length < *buffersize) {
			buffer [length++] = (u_int8_t)result;
		} else {
                     DT((DT_AM, "too short buffer in cnf_get_hex_buf\n"));
                     exit(1);
		}
		
		if (!*next){
                     break;
		}
		
		line = next;
	}
	
	*buffersize = length;
}

void config_sections (FILE* file, Section sections [])
{
	Section* section;
	
	for (section = sections; section->name; section++){
		cnf_sect (file, section);
	}
}
