/***************************************************************************
 * The contents of this file are subject to the NOKOS License Version 1.0a
 * (the "License"); you may not use this file except in compliance with the
 * License. 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF  ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The Original Software is Nokia D211 Software for Linux.
 *
 * Copyright (c) 2002 Nokia. All Rights Reserved.
 *
 * Nokia is a registered trademark of Nokia Corporation.
 * Other product and company names mentioned herein may be
 * trademarks or tradenames of their respective owners. 
 *
 * Contributor(s): ______________________________________. 
 *
 **************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include "naap_am.h"
#include "naap_socket.h"
#include "naap_dt.h"
#include "am_socket.h"
#include "naap_memory.h"
#include "naap_event.h"

int get_socket_descriptor(Socket_handle s)
{
  return s->fd;
}

Socket_handle am_create_udp_socket(IP_addr addr, u_int16_t *port)
{
  Socket_handle_impl* s;
  int result;
  struct sockaddr_in saddr;
  const int false = 0;
  const int true = 1;
  struct in_addr mcast_addr;

  DT((DT_SOCKET, "am_create_udp_socket: addr %s, port %d\n",
      am_inet_ntoa(addr), ntohs(*port)));

  s = (Socket_handle_impl *)am_alloc(sizeof(Socket_handle_impl));

  s->fd = socket (AF_INET, SOCK_DGRAM, 0); 

  if (s->fd < 0)
    {
      DT((DT_ERROR, "ERROR: am_create_udp_socket: socket failed, error %s\n",
	  strerror(errno)));
      am_free(s);
      return INVALID_NAAP_SOCKET;
    }

  /* Only one agent instance at a time can bind a socket. */
  result = setsockopt (s->fd, SOL_SOCKET, SO_REUSEADDR, (int*)&false,
		       sizeof(false));

  if (result < 0){
    DT((DT_SOCKET, "am_create_udp_socket: setsockopt failed: %s\n",
	strerror(errno)));
    am_free(s);
    return INVALID_NAAP_SOCKET;
  }


  saddr.sin_family = AF_INET;
  saddr.sin_addr.s_addr = addr;
  
  saddr.sin_port = *port;


  result = bind (s->fd, (struct sockaddr*) &saddr, (sizeof (saddr)));
	
  if (result < 0){
    DT((DT_SOCKET, "am_create_udp_socket: bind failed: %s\n",
	strerror(errno)));
    am_free(s);
    return INVALID_NAAP_SOCKET;
  }

  mcast_addr.s_addr = addr;

  result = setsockopt(s->fd, IPPROTO_IP, IP_MULTICAST_IF,
		      (u_int8_t *)&mcast_addr, sizeof(struct in_addr));

  if (result < 0)
    {
      DT((DT_ERROR,
	  "am_create_udp_socket: setsockopt(IP_MULTICAST_IF) failed, %s\n",
	  strerror(errno)));
      am_free(s);
      return INVALID_NAAP_SOCKET;
    }
  
  
  result = setsockopt(s->fd, SOL_SOCKET, SO_BROADCAST,
		      (int*)&true, sizeof(true));
  
  if (result < 0)
    {
      DT((DT_ERROR,
	  "am_create_udp_socket: setsockopt(SO_BROADCAST) failed, %s\n",
	  strerror(errno)));
      free(s);
      return INVALID_NAAP_SOCKET;
    }

  return s;
}

int am_sendto(Socket_handle s, u_int8_t *data, int datalen,
	      IP_addr dest_addr, u_int16_t dest_port)
{
  int result;
  struct sockaddr_in addr;

  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = dest_addr;
  addr.sin_port = dest_port;
  
  result = sendto(
		  get_socket_descriptor(s),
		  (char *)data,
		  datalen,
		  0,
		  (struct sockaddr*) &addr,
		  sizeof (struct sockaddr_in));
  
  
  if (result < 0) 
    {
      DT((DT_SOCKET,
	  "ERROR: am_sendto(%s:%d) sendto failed: %s\n",
	  am_inet_ntoa(dest_addr), am_htons(dest_port), strerror(errno)));
      return -1;
    }

  return result;
}

int am_recvfrom(Socket_handle s, u_int8_t *buf, int buflen,
		IP_addr* source_addr, u_int16_t* source_port)
{

  struct sockaddr_in from;
  int addrlen = sizeof (from);
  int bytes;

  bytes = recvfrom(
		   get_socket_descriptor(s),
		   (char *)buf,
		   buflen,
		   0,
		   (struct sockaddr*)&from,
		   &addrlen);
	
  if (bytes < 0)
    {
      DT((DT_SOCKET,
	  "ERROR: am_recvfrom: recvfrom failed: %s\n",
	  strerror(errno)));
      return -1;
    }

  *source_addr = from.sin_addr.s_addr;
  *source_port = from.sin_port;
	
  return bytes;
}


int am_join_multicast_group(Socket_handle s, 
			    IP_addr local_addr,
			    IP_addr group_addr)
{
  struct ip_mreq req;
  int result;
  int sock = get_socket_descriptor(s);

  req.imr_multiaddr.s_addr = group_addr;
  req.imr_interface.s_addr = local_addr;

  result = setsockopt(sock, IPPROTO_IP,
		      IP_ADD_MEMBERSHIP, (char *)&req, sizeof (req));
	
  if (result < 0)
    {
      DT((DT_SOCKET,
	  "ERROR: am_join_multicast_group: IP_ADD_MEMBERSHIP failed\n"));
      return 1;
    } 

  return 0;
}


int am_leave_multicast_group(Socket_handle s,
			     IP_addr local_addr,
			     IP_addr group_addr)
{

  struct ip_mreq req;
  int result;
  int sock = get_socket_descriptor(s);
	
  req.imr_multiaddr.s_addr = group_addr;
  req.imr_interface.s_addr = local_addr;

  result = setsockopt (sock, IPPROTO_IP,
		       IP_DROP_MEMBERSHIP, (char *)&req, sizeof (req));
	
  if (result < 0)
    {
      DT((DT_SOCKET, "ERROR: IP_DROP_MEMBERSHIP failed\n"));
      return 1;
    } 
  
  return 0;
}

int am_close_socket(Socket_handle s)
{
  int ret;
  
  ret = close(get_socket_descriptor(s));
  free(s);
  
  return ret;
}

int am_is_mcast_addr (IP_addr addr)
{
  return IN_MULTICAST(ntohl(addr));
}


IP_addr am_inet_aton(u_int8_t *ip_addr)
{
  IP_addr ret;

  ret = inet_addr(ip_addr);

  if (ret == INADDR_NONE) 
    {
      return IP_ADDR_NONE;
    } 
  else 
    {
      return ret;
    }

}


u_int8_t* am_inet_ntoa(IP_addr addr)
{
  struct in_addr tmp;

  tmp.s_addr = addr;

  return inet_ntoa(tmp);
}


u_int16_t am_htons(u_int16_t hostshort)
{
  return htons(hostshort);
}


u_int32_t am_htonl(u_int32_t hostlong)
{
  return htonl(hostlong);
}


u_int16_t am_ntohs(u_int16_t netshort)
{
  return ntohs(netshort);
}


u_int32_t am_ntohl(u_int32_t netlong)
{
  return ntohl(netlong);
}

