/***************************************************************************
 * The contents of this file are subject to the NOKOS License Version 1.0a
 * (the "License"); you may not use this file except in compliance with the
 * License. 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF  ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the 
 * License. 
 *
 * The Original Software is Nokia D211 Software for Linux.
 *
 * Copyright (c) 2002 Nokia. All Rights Reserved.
 *
 * Nokia is a registered trademark of Nokia Corporation.
 * Other product and company names mentioned herein may be
 * trademarks or tradenames of their respective owners. 
 *
 * Contributor(s): ______________________________________. 
 *
 **************************************************************************/

#ifndef NAAP_SOCKET_H
#define NAAP_SOCKET_H

#include "naap_am.h"

#ifdef __cplusplus
extern "C" {
#endif
	

/* store addresses in an IP_addr in network byte order!*/
typedef u_int32_t IP_addr; 

#define IP_ADDR_ANY (IP_addr)0x00000000
#define IP_ADDR_NONE (IP_addr)0xffffffff
#define IP_ADDR_BROADCAST (IP_addr)0xffffffff


/* Socket_handles are opaque to the NAAP core. The core gets
   them from the AM and uses them only as parameters to adaptation routine
   calls. The AM must provide implementation for struct Socket_handle_impl 

*/

typedef struct Socket_handle_impl* Socket_handle;

#define INVALID_NAAP_SOCKET 0


/* The adaptation module calls nc_handle_socket when there is 
   data available for reading from socket s with am_recvfrom
   Used on: NC, NA */
void nc_handle_socket(Socket_handle s);

/* nc_get_receiving_sockets returns a pointer to a Socket_handle table 
   with "count" members. These are the sockets that the NAAP core is using for 
   receiving data.
   The adaptation module must check if there is anything to read
   from these sockets and if there is, the AM calls nc_handle_socket 
   Used on: NC, NA */
Socket_handle* nc_get_receiving_sockets (int32_t* count);


/* Create a UDP Socket_handle 
   and bind the socket to a local IP address and a port

   If addr is IP_ADDR_ANY, the AM may pick any local address.

   if argument *port == 0, an available port number is selected and
   written to *port. If *port != 0, the socket is bound to *port.
   Give port in network byte order.

   Socket_handles are always non-blocking

   return value: a Socket_handle if successful
                 INVALID_MIPAI_SOCKET if unsuccessful 

   Used on: NC, NA */

Socket_handle am_create_udp_socket(IP_addr address, u_int16_t* port);


/* Close a Socket_handle
   return value: 0 if successful 

   Used on: NC, NA */
int32_t am_close_socket(Socket_handle s);

/* Send data to the network.
   Give dest_port in network byte order.
   returns the total number of bytes sent or -1 if error

   The NAAP Core sends UDP packets by unicast and multicast.
   The adaptation module takes care of enabling the device's multicast 
   functionality. 
   The adaptation module must make sure that the network interface for 
   outgoing multicast packets on NAAP client is correct

   Used on: NC, NA */
int32_t am_sendto(Socket_handle s, u_int8_t *data, int32_t datalen,
	      IP_addr dest_addr, u_int16_t dest_port);


/* Receive data from the network and store the source address 
   and the source port.
   source_port is given in network byte order.
   returns the total number of bytes received  or -1 if error

   Used on: NC, NA */
int32_t am_recvfrom(Socket_handle s, u_int8_t *buf, int32_t buflen,
		IP_addr* source_addr, u_int16_t* source_port);


/* check wheter addr is a multicast address 
   Used on: NC, NA */
int32_t am_is_mcast_addr (IP_addr addr);

/* Convert a dotted IP address string to IP_addr 
   if ip_addr cannot be converted, the function returns IP_ADDR_NONE 
   Used on: NC, NA */
IP_addr am_inet_aton(u_int8_t *ip_addr);

u_int8_t* am_inet_ntoa(IP_addr);

/* Convert from host byte order to network byte order 
   and vice versa 
   Used on: NC, NA */
u_int16_t am_htons(u_int16_t hostshort);
u_int32_t am_htonl(u_int32_t hostlong);

u_int16_t am_ntohs(u_int16_t netshort);
u_int32_t am_ntohl(u_int32_t netlong);



#ifdef __cplusplus
}
#endif

#endif /* NAAP_SOCKET_H not defined */
