/***************************************************************************
 * The contents of this file are subject to the NOKOS License Version 1.0a
 * (the "License"); you may not use this file except in compliance with the
 * License. 
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF  ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the 
 * License. 
 * 
 * The Original Software is Nokia D211 Software for Linux. 
 *
 * Copyright (c) Nokia 2002 and others. All Rights Reserved.
 *
 * Nokia is a registered trademark of Nokia Corporation.
 * Other product and company names mentioned herein may be
 * trademarks or tradenames of their respective owners. 
 *
 * Contributor(s): ______________________________________. 
 *
 **************************************************************************/

/***************************************************************************
 *
 * Virtual serialport implementation
 *
 **************************************************************************/

#include "nokia_info.h"
#include "nokia_priv.h"
#include "dserial_priv.h"

/* Number of serial ports */
#define SERIAL_D211_NR     1

/*
 * Things needed by tty driver
 */
static struct tty_driver dnormal_driver;
/*static int dser_refcount = 0; */
static struct tty_struct *dser_table[SERIAL_D211_NR];
static struct termios *dser_termios[SERIAL_D211_NR];
static struct termios *dser_termios_locked[SERIAL_D211_NR];

static int dser_set_signals(dser_t* dser, unsigned int cmd, unsigned int new);

/*
 * Driver internals
 */
static dser_t *saved_dser[SERIAL_D211_NR];

/*
 * debug stuff
 */
#define DEBUG_ALL
#ifdef DEBUG_ALL
#define DEBUG_FUNC  DEBUG("%s\n", __FUNCTION__);
#define DEBUG_FUNC_EXIT  DEBUG("%s exit\n", __FUNCTION__);
#define DEBUG_ERR(args...) ERROR(args)
#else
#define DEBUG_FUNC
#define DEBUG_FUNC_EXIT  
#define DEBUG_ERR(args...) dprintk (args)
#endif

/*
 *      This routine is called when tty device is opened.
 */
static int dser_open(struct tty_struct * tty, struct file * filp){
        dser_t *dser;
        int line;
        unsigned long flags;
        
        DEBUG_FUNC;

        save_flags(flags);cli();

        /* MDU: count managed from outside MOD_INC_USE_COUNT; */
	try_module_get(THIS_MODULE);

        if (!tty){
                /* MDU: count managed from outside MOD_DEC_USE_COUNT; */
		module_put(THIS_MODULE);
                restore_flags(flags);
                DEBUG("tty NULL\n");
                return -ENODEV;
        }
        
        /* MDU not sure about this one!!!!!!! 
	line = MINOR(tty->device) - tty->driver.minor_start; */
	line = tty->index;

	ERROR("Opening device %i\n", line);
        DEBUG("Opening device %i\n", line);

        if (line >= SERIAL_D211_NR) {
                /* MDU: count managed from outside MOD_DEC_USE_COUNT; */
		module_put(THIS_MODULE);
                restore_flags(flags);
                DEBUG("ENODEV\n");
                return -ENODEV;
        }
        
        dser = saved_dser[line];
        if (!dser){
                /* MDU: count managed from outside MOD_DEC_USE_COUNT; */
		module_put(THIS_MODULE);
                restore_flags(flags);
                DEBUG("dser NULL\n");
                return -ENODEV;
        }

        spin_lock_bh(&dser->lock);

        dser->count++;

        tty->driver_data = dser;
        dser->tty = tty;
        tty->low_latency = (dser->flags & ASYNC_LOW_LATENCY) ? 1 : 0;
        
        spin_unlock_bh(&dser->lock);

        restore_flags(flags);

        dser_set_signals(dser, TIOCMBIS, TIOCM_RTS | TIOCM_DTR);        

        DEBUG_FUNC_EXIT;
        return 0;
}

extern wait_queue_head_t stop_queue;
/*
 *      This routine is called when tty device is closed.
 */
static void dser_close(struct tty_struct * tty, struct file * filp){
        dser_t *dser;
        int line;
        unsigned long flags;

        DEBUG_FUNC;

        save_flags(flags);cli();

        if (!tty){
                ERROR("No TTY\n");
                restore_flags(flags);
                return;
        }

        /* MDU not sure about this one!!!!!!! 
	line = MINOR(tty->device) - tty->driver.minor_start; */
	line = tty->index;

        ERROR("Closing device %i\n", line);
        DEBUG("Closing device %i\n", line);
        
        dser = saved_dser[line];
        if (!dser){
                ERROR("No dser\n");
                restore_flags(flags);
                return;
        }

        spin_lock_bh(&dser->lock);

        dser->count--;

        if( dser->count ) {      /* do nothing */
                DEBUG("driver count is not null\n");
                /* MDU: count managed from outside MOD_DEC_USE_COUNT; */
		module_put(THIS_MODULE);
                spin_unlock_bh(&dser->lock);
                restore_flags(flags);
                wake_up_interruptible(&stop_queue);
                return; 
        }

        /*
         * Now we wait for the transmit buffer to clear; and we notify 
         * the line discipline to only process XON/XOFF characters.
         */
        tty->closing = 1;
        DEBUG("calling tty_wait_until_sent()\n");
        tty_wait_until_sent(tty, 10*HZ);


        /* drop ldisc's buffer */
        if (tty->ldisc.flush_buffer)
                tty->ldisc.flush_buffer(tty);

	/* MDU 
        if (tty->driver.flush_buffer) 
                tty->driver.flush_buffer(dser->tty); */  
        if (tty->driver->flush_buffer) 
                tty->driver->flush_buffer(dser->tty);

        tty->closing = 0;
        tty->driver_data = NULL;
        dser->tty = NULL;
        
        /* MDU: count managed from outside MOD_DEC_USE_COUNT; */
	module_put(THIS_MODULE);

        spin_unlock_bh(&dser->lock);
        
        restore_flags(flags);
        
        wake_up_interruptible(&stop_queue);

        dser_set_signals(dser, TIOCMBIC, TIOCM_RTS | TIOCM_DTR);
        
        DEBUG_FUNC_EXIT;
        return;
}

/*
 *      This routine is called by the kernel to write a series of
 *      characters to the tty device.  The characters may come from
 *      user space or kernel space.  This routine will return the
 *      number of characters actually accepted for writing.
 */
static int dser_write(struct tty_struct * tty, int from_user,
                   const unsigned char *buf, int count){

        dser_t* dser;
        driver_priv_t *priv;
        char tmp_buf[SERIAL_XMIT_MAX_SIZE];
        int len, ret = 0;

        DEBUG_FUNC;

        if (!tty || !buf || !count)
                return 0;
        
        dser = tty->driver_data;
        if (!dser || !dser->priv)
                return 0;
        priv = dser->priv;
        
        while(count > 0){
                if (count < SERIAL_XMIT_MAX_SIZE)
                        len = count;
                else
                        len = SERIAL_XMIT_MAX_SIZE;
                
                if (from_user) { //buffer from userspace
                        len -= copy_from_user(tmp_buf, &buf[ret], len);
                        if (!len) { //no data copied
                                if (!ret)
                                        ret = -EFAULT;
                                break;
                        }
                } else {
                        memcpy(tmp_buf, &buf[ret], len);
                }
                
#ifdef DEBUG_FRAMES
                DEBUG ("Send Buffer (%i):", len);                  
                {
                        int i;
                        for (i=0; i<len; i++){
                                dprintk ("%x ", 0xff&tmp_buf[i]);
                        }
                }
                dprintk ("\n");
#endif
                spin_lock_bh(&dser->lock);

                if (!(dser->msig & TIOCM_CTS)){
                        DEBUG("dmgr_serial_send failed: No CTS\n");
                        spin_unlock_bh(&dser->lock);
                        break;
                }
                spin_unlock_bh(&dser->lock);
               
                if ((len = dnc_serial_send(priv, tmp_buf, len)) <= 0){
                        DEBUG("dmgr_serial_send failed\n");
                        break;
                }
                
                ret += len;
                count -= len;
        }

        spin_lock_bh(&dser->lock);

        dser->tx_bytes += ret;
        if (count > ret)
                dser->tx_error_bytes += count - ret;

        spin_unlock_bh(&dser->lock);
                
        DEBUG_FUNC_EXIT;
        return ret;
}

/*
 *      This routine returns the numbers of characters the tty driver
 *      will accept for queuing to be written.  This number is subject
 *      to change as output buffers get emptied, or if the output flow
 *      control is acted.
 */
static int  dser_write_room(struct tty_struct *tty){
        dser_t* dser;
        int ret = 0;

        DEBUG_FUNC;
        if (!tty)
                return 0;
        
        dser = tty->driver_data;
        if (!dser)
                return 0;

        spin_lock_bh(&dser->lock);
        
        if (dser->msig & TIOCM_CTS){
                ret = SERIAL_XMIT_MAX_SIZE;
        }
        
        spin_unlock_bh(&dser->lock);

        DEBUG_FUNC_EXIT;
        
        return ret;
}


/*
 *      This routine notifies the tty driver that it should stop
 *      outputting characters to the tty device.  
 */
static void dser_stop(struct tty_struct *tty){
        DEBUG_FUNC;

        DEBUG_FUNC_EXIT; 
}

/*
 *      This routine notifies the tty driver that it resume sending
 *      characters to the tty device.
 */
static void dser_start(struct tty_struct *tty){
        DEBUG_FUNC;
 
        DEBUG_FUNC_EXIT; 
}

/*
 *      This routine set modem signals.
 */
static int dser_set_signals(dser_t* dser, unsigned int cmd, unsigned int new){
        unsigned int signal = 0;

        spin_lock_bh(&dser->lock);
                
        if (cmd == TIOCMBIS)
                dser->msig |= new;
        else if (cmd == TIOCMBIC)
                dser->msig &= ~new;
        else if (cmd == TIOCMSET)
                dser->msig = new;
        else {
                spin_unlock_bh(&dser->lock);
                return -EINVAL;
        }
        
        if(dser->msig & TIOCM_DTR)
                signal |= MSIG_DTR;
        if(dser->msig & TIOCM_RTS)
                signal |= MSIG_RTS;
        
        dnc_serial_signal(dser->priv, signal);

        spin_unlock_bh(&dser->lock);
        return 0;
}


/*
 *      This routine allows the tty driver to implement
 *      device-specific ioctl's.  If the ioctl number passed in cmd
 *      is not recognized by the driver, it should return ENOIOCTLCMD.
 */
static int  dser_ioctl(struct tty_struct *tty, struct file * file,
                    unsigned int cmd, unsigned long arg){
        unsigned int ret, new;
        dser_t* dser;
        DEBUG_FUNC;

        if (!tty)
                return 0;
        
        dser = tty->driver_data;
        if (!dser)
                return 0;

	switch (cmd) {
        case TIOCMGET:
                spin_lock_bh(&dser->lock);
                ret = put_user(dser->msig, (unsigned int*) arg);
                spin_unlock_bh(&dser->lock);
                return ret;
        case TIOCMBIS:
        case TIOCMBIC:
        case TIOCMSET:
                if (get_user(new, (unsigned int*)arg))
                        return -EFAULT;
                return dser_set_signals(dser, cmd, new);
                /*case TIOCGSERIAL:
                case TIOCSSERIAL:
                case TIOCSERGETLSR: 
                case TIOCMIWAIT:
                case TIOCGICOUNT: */
        default:
                DEBUG("Unknown IOCTL 0x%04x\n", cmd);
                return -ENOIOCTLCMD;
	}
        DEBUG_FUNC_EXIT; 
        return -ENOIOCTLCMD;
}

/*
 * Returns characters in buffer
 */
static int  dser_chars_in_buffer(struct tty_struct *tty){
        return 0;
}


/*
 * Receive function, called when serial data is received
 */

void dser_receive(driver_priv_t *priv, char* buf, int count){

        struct tty_struct *tty;
        dser_t *dser;
        int len;

        DEBUG_FUNC;

        if (!priv || !buf || !count){
                ERROR("priv/buf/count NULL");
                return;
        }
        dser = priv->ser;
        if (!dser){
                ERROR("dser NULL");
                return;
        }
        spin_lock_bh(&dser->lock);
        
        tty = dser->tty;
        if (!tty){
                spin_unlock_bh(&dser->lock);
                DEBUG("tty NULL");
                return;
        }

        len = count;
        //check that there is enough space in the buffer
        if ((tty->flip.count + count) >= TTY_FLIPBUF_SIZE){
                //Not enough room in tty buffer, trying to add some
                /* MDU: task queue has changed to workqueue
								tty->flip.tqueue.routine((void *)tty); */
		tty->flip.work.func((void *)tty);

                if ((tty->flip.count + count) >= TTY_FLIPBUF_SIZE) {
                        //max length that can be copied now
                        len = TTY_FLIPBUF_SIZE - tty->flip.count;

                        DEBUG_ERR ("TTY_DONT_FLIP set, %i bytes of rx data lost\n", count-len);
                        dser->rx_error_bytes += count - len;
                }
        }
#ifdef DEBUG_FRAMES
        DEBUG ("Recv Buffer (%i): ", len);
        {
                int i;
                for (i=0; i<len; i++){
                        dprintk ("%x ", 0xff&buf[i]);
                }
        }
        dprintk ("\n");
#endif

        if (len > 0){
                //copy data
                memcpy(tty->flip.char_buf_ptr, buf, len);
                memset(tty->flip.flag_buf_ptr, TTY_NORMAL, len);
                
                //update pointers and statistic
                tty->flip.flag_buf_ptr += len;
                tty->flip.char_buf_ptr += len;
                tty->flip.count += len;
                dser->rx_bytes += len;
        }

        tty_flip_buffer_push(tty);

        spin_unlock_bh(&dser->lock);

        DEBUG_FUNC_EXIT;
                
        return;
}

void dser_tty_hangup(driver_priv_t *priv){
        if (!priv || !priv->ser || !priv->ser->tty) return;

        DEBUG ("Hangup tty\n");
        tty_hangup(priv->ser->tty);
}
        

/*
 * Wake up tty
 */
void dser_wake_up_write(driver_priv_t *priv){        

        if (!priv || !priv->ser || !priv->ser->tty) return;

        DEBUG_FUNC;

        spin_lock_bh(&priv->ser->lock);

        if ((priv->ser->tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
            priv->ser->tty->ldisc.write_wakeup)
                (priv->ser->tty->ldisc.write_wakeup)(priv->ser->tty);

        wake_up_interruptible(&priv->ser->tty->write_wait);
        
        spin_unlock_bh(&priv->ser->lock);

        DEBUG_FUNC_EXIT;
        
        return;
}

/*
 * Receive modem signals
 */
void dser_signal(driver_priv_t *priv, int modem_signal){

        DEBUG_FUNC;

        if (!priv || !priv->ser) return; 
        /* store modem signals */
        spin_lock_bh(&priv->ser->lock);

        if(modem_signal & MSIG_CTS)
                priv->ser->msig |= TIOCM_CTS;
        else
                priv->ser->msig &= ~TIOCM_CTS;
        if(modem_signal & MSIG_DCD)
                priv->ser->msig |= TIOCM_CD;
        else
                priv->ser->msig &= ~TIOCM_CD;
        if(modem_signal & MSIG_DSR)
                priv->ser->msig |= TIOCM_DSR;
        else
                priv->ser->msig &= ~TIOCM_DSR;
        
        DEBUG ("Signal state: 0x%x\n", priv->ser->msig);
        if (priv->ser->msig & TIOCM_CTS){
                spin_unlock_bh(&priv->ser->lock);
                dser_wake_up_write(priv);
        }
        else
                spin_unlock_bh(&priv->ser->lock);

        DEBUG_FUNC_EXIT;
}


/*
 * Gets statistics
 */
void dser_get_stats(driver_priv_t *priv, dser_stats_t *stats){
        if (!priv || !stats || !priv->ser)
                return;
        spin_lock_bh(&priv->ser->lock);
        stats->rx_bytes = priv->ser->rx_bytes;
        stats->tx_bytes = priv->ser->tx_bytes;
        stats->rx_error_bytes = priv->ser->rx_error_bytes;
        stats->tx_error_bytes = priv->ser->tx_error_bytes;
        spin_unlock_bh(&priv->ser->lock);
}

/*
 * Initialization function
 */

dser_t *dser_new(driver_priv_t *priv){
        dser_t *dser;
        int index;
        
        DEBUG_FUNC;

        if(!priv) return NULL;
        
        for(index=0; index < SERIAL_D211_NR; index++) {
                if(saved_dser[index] == NULL) break;
        }
        if(index == SERIAL_D211_NR) {
                ERROR("SERIAL_D211_NR count exceeded! Can't open more devices.\n");
                return NULL;
        }
                
        //allocate dser_t structure
        dser = kmalloc(sizeof(dser_t), GFP_KERNEL);
        if(!dser) goto dser_new_fail;
        memset(dser,0,sizeof(dser_t));

        memset(&dnormal_driver,0,sizeof(struct tty_driver));

        //initialize dser
        dser->flags = 0; //control tty, checked in open
        dser->msig = 0; 
        spin_lock_init(&dser->lock);
        spin_lock_bh(&dser->lock);
                       
        //initialize driver        
        dnormal_driver.magic = TTY_DRIVER_MAGIC;
        dnormal_driver.driver_name = "serial_d211";
        dnormal_driver.name = SERIAL_D211_NAME;
        dnormal_driver.major = SERIAL_D211_MAJOR;
        dnormal_driver.minor_start = SERIAL_D211_MINOR;
        dnormal_driver.num = SERIAL_D211_NR;
        dnormal_driver.type = TTY_DRIVER_TYPE_SERIAL;
        dnormal_driver.subtype = SERIAL_TYPE_NORMAL;
        dnormal_driver.init_termios = tty_std_termios;
        dnormal_driver.flags = TTY_DRIVER_REAL_RAW | TTY_DRIVER_NO_DEVFS;
        /* MDU refcount is already set to 0 by memset 
	 * dnormal_driver.refcount = &dser_refcount; */
        /* MDU dnormal_driver.table = dser_table; */
        dnormal_driver.ttys = dser_table;
        dnormal_driver.termios = dser_termios;
        dnormal_driver.termios_locked = dser_termios_locked;

        //set function pointers (others are NULLed with memset above)
        dnormal_driver.open = dser_open;
        dnormal_driver.close = dser_close;
        dnormal_driver.write = dser_write;
        dnormal_driver.ioctl = dser_ioctl;
        dnormal_driver.write_room = dser_write_room;
        dnormal_driver.stop = dser_stop;
        dnormal_driver.start = dser_start;
        dnormal_driver.chars_in_buffer = dser_chars_in_buffer;
        
        //register driver
        if (tty_register_driver(&dnormal_driver)){
                DEBUG_ERR("Couldn't register D211 serial driver\n");
                spin_unlock_bh(&dser->lock);
                goto dser_new_fail;
        }
        else
                DEBUG("TTY driver registered\n");

        saved_dser[index] = dser;

        dser->priv = priv;

        spin_unlock_bh(&dser->lock);

        DEBUG_FUNC_EXIT; 
        
        return dser;

  dser_new_fail:

        if(dser) {
                dser_delete(dser);
        }
        return NULL;
}

/*
 * Deinitialization function
 */
void dser_delete(dser_t *dser)
{
        int index, ret;

        DEBUG_FUNC;
        
        if(!dser) return;

        spin_lock_bh(&dser->lock);

        for(index = 0; index < SERIAL_D211_NR; index++) {
                if(saved_dser[index] == dser){
                        saved_dser[index] = NULL;
                                //deregister driver
                        if ((ret = tty_unregister_driver(&dnormal_driver)))
                                DEBUG_ERR("Couldn't unregister serial driver: %i\n", ret);
                        else
                                DEBUG("TTY driver unregistered\n");

                }
        }

        spin_unlock_bh(&dser->lock);

        dfree(dser, sizeof(dser_t));
        DEBUG_FUNC_EXIT; 
        return;
}




























